// anydll.cpp : Definiert den Einstiegspunkt fr die Konsolenanwendung.
//

#include "stdafx.h"

struct parameter_t {
	_TCHAR *value;

	enum type_t {
		tString,
		tDWord,
		tDouble,
		tUnknown
	} type;
};

struct option_t {
	enum charset_t {
		ascii,
		wide
	} charset;
} option;

struct errorValue_t {
	enum {
		ok = 0,
		unknown_option,
		number_parameters,
		load_library,
		access_proc
	};
};

typedef std::vector<parameter_t> parameterList;

void findOutType(parameter_t& param)
{
	std::basic_stringstream<_TCHAR> tss;
	std::basic_string<_TCHAR> tbs;

	DWORD dwValue = _ttoi(param.value);
	tss << dwValue;
	tss >> tbs;

	if (_tcscmp(param.value, tbs.c_str()) == 0) {
		param.type = parameter_t::tDWord;
		return;
	}

	param.type = parameter_t::tString;
}

void callDll(parameterList& parameter, int nParam)
{
	HMODULE hLib = LoadLibrary(parameter[0].value);
	if (hLib == NULL) {
		std::basic_string<_TCHAR> libName = parameter[0].value;
#ifdef _UNICODE
		libName += L".dll";
#else
		libName += ".dll";
#endif
		hLib = LoadLibrary(libName.c_str());
		if (hLib == NULL) {
			std::cerr << "Cannot load library\n";
			exit(errorValue_t::load_library);
		}
	}

	FARPROC proc = GetProcAddress(hLib, CT2A(parameter[1].value));
	if (proc == NULL) {
		std::string procName = (LPSTR) CT2A(parameter[1].value);
		if (option.charset == option_t::ascii)
			procName += "A";
		else
			procName += "W";

		proc = GetProcAddress(hLib, procName.c_str());
		if (proc == NULL) {
			std::cerr << "Cannot access procedure\n";
			exit(errorValue_t::access_proc);
		}
	}

	for (int i=2; i<nParam; i++)
		if (parameter[i].type == parameter_t::tUnknown)
			findOutType(parameter[i]);

	CT2W *pw;
	CT2A *pa;

	CHAR *asciiString;
	WCHAR *wideString;

	int retVal;

	if (nParam <= 2)
		goto call_proc;

	int i = nParam - 1;

	unsigned int stack_pointer;

	__asm mov stack_pointer, esp

begin_loop:
	if (parameter[i].type == parameter_t::tDWord)
		goto type_dword;

	if (parameter[i].type == parameter_t::tDouble)
		goto type_double;

// type_string:
	if (option.charset == option_t::wide)
		goto string_wide;

// string_ascii:
	pa = new CT2A(parameter[i].value);
	asciiString = *pa;
	__asm {
		mov  eax, asciiString;
		push eax
	}
	goto end_loop;

string_wide:
	pw = new CT2W(parameter[i].value);
	wideString = parameter[i].value;
	__asm {
		mov  eax, wideString;
		push eax
	}
	goto end_loop;

type_dword:
	DWORD dwValue = _tstoi(parameter[i].value);
	__asm {
		mov eax, dwValue
		push eax
	}
	goto end_loop;

type_double:
	double dValue = _tstof(parameter[i].value);

	DWORD doubleFirst  = ((DWORD *) &dValue)[0];
	DWORD doubleSecond = ((DWORD *) &dValue)[1];

	__asm {
		mov eax, doubleFirst
		push eax
		mov eax, doubleSecond
		push eax
	}

end_loop:
	if (--i >= 2)
		goto begin_loop;

call_proc:
	__asm call proc

// the_end:
	__asm mov retVal, eax
	__asm mov esp, stack_pointer

	std::cout << retVal << "\n";
}

int _tmain(int argc, _TCHAR* argv[])
{
	option.charset = option_t::wide;
	parameterList parameter;
	parameter_t::type_t actualType = parameter_t::tUnknown;
	int nParam = 0, i;

	for (i=1; i<argc; i++) {
		switch (argv[i][0]) {
			case '/':
			case '-':
				switch (argv[i][1]) {
				case '?':
				case 'h':
				case 'H':
					std::cout << "Calls an arbitrary function in an arbitrary DLL and prints the return value.\n\n";
					std::cout << "ANYDLL [/A|/W] DLLName ProcName [[/S|/D|/O] Param1 [[/S|/D|/O] Param2 [...]]]\n\n";
					std::cout << "  DllName\tName of the DLL, where the function should be loaded from\n";
					std::cout << "  ProcName\tName of the function to call\n";
					std::cout << "  /A\t\tParameters are sent 8bit encoded\n";
					std::cout << "  /W\t\tParameters are sent UNICODE (default)\n";
					std::cout << "  /S\t\tNext parameter is PTSTR\n";
					std::cout << "  /D\t\tNext parameter is DWORD\n";
					std::cout << "  /O\t\tNext parameter is double\n";
					std::cout << "\nCopyleft (c) 2006, Volker Birk\n";
					std::cout << "This program is under the GNU General Public License 2.0.\n";
					exit(0);

				case 'a':
				case 'A':
					option.charset = option_t::ascii;
					break;

				case 'w':
				case 'W':
					option.charset = option_t::wide;
					break;

				case 's':
				case 'S':
					actualType = parameter_t::tString;
					break;

				case 'd':
				case 'D':
					actualType = parameter_t::tDWord;
					break;

				case 'o':
				case 'O':
					actualType = parameter_t::tDouble;
					break;

				default:
					std::cerr << "Unknown option: " << argv[i][1] << "\n";
					exit(errorValue_t::unknown_option);
				}
				break;

			default:
				parameter_t newParam;
				newParam.value = argv[i];
				newParam.type = actualType;

				parameter.push_back(newParam);
				nParam++;

				actualType = parameter_t::tUnknown;
		}
	}
	if (nParam < 2) {
		std::cerr << "Illegal number of parameters\n";
		exit(errorValue_t::number_parameters);
	}
	
	callDll(parameter, nParam);

	return errorValue_t::ok;
}
